/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * dataset - functions for managing data sets. Usually one instance
 * represents all the data loaded from a single data file.
 */

#include <ctype.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "dataset.xh"
#include "stringutil.xh"
#include "system.xh"
#include "util.xh"

static void free_dataset(my_dataset_t *dataset);

static my_time_t parse_timestamp(const char **p) {
   return (my_time_t)(parse_double_num(p) * TIME_MULTIPLIER);
}

static void parse_line(my_dataset_t *dataset, char *line, my_time_t timeoff, boolean_t is_millis) {
   const char *p = line;

   my_time_t timestamp = parse_timestamp(&p);
   if(!*p) {
      //only timestamp or nothing at all
      return;
   }
   if(!my_isspace(*p)) {
      chomp(line);
      fprintf(stderr, "Bad data '%s' in line '%s'\n", p, line);
      return;
   }

   if(is_millis) {
      timestamp /= 1000;
   }

   timestamp += timeoff;

   const int lineidx = dataset->time_count;

   my_time_t cur_time_diff = 0;
   if(likely(lineidx > 0)) {
      cur_time_diff = timestamp - dataset->time[lineidx - 1];
      if(unlikely(cur_time_diff <= 0)) {
	 chomp(line);
	 fprintf(stderr, "Time goes back or does not advance (prev timestamps %Ld, %Ld) - ignoring line '%s'\n",
		 dataset->time[lineidx - 2], dataset->time[lineidx - 1], line);
	 return;
      }
   }

   my_time_arr_ensure(dataset->time, lineidx + 1);

   dataset->time[lineidx] = timestamp + timeoff;

   if(likely(lineidx > 0)) {
      if(likely(cur_time_diff > 0) && (cur_time_diff < dataset->min_time_diff || unlikely(!dataset->min_time_diff))) {
	 dataset->min_time_diff = cur_time_diff;
      }
   }
   
   double prev[100]; //##TODO## not fixed
   int i;
   for(i=0; *(p = skipws(p)); ++i) {
      my_data_t num;
      do {
	 num = parse_double_num(&p);
	 if(!my_isspace(*p) && *p) {
	    // failed to parse number, check quickly why
	    const char *pe = skipnonws(p);
	    if(*p != ':') {
	       // it's not a time field.. complain (if it is a timefield, don't complain, just ignore)
	       chomp(line);
	       fprintf(stderr, "Ignoring bad data '\033[1m%.*s\033[m' in line:\n  %.*s\033[1m%.*s\033[m%s\n\n", (int)(pe-p), p, (int)(p-line), line, (int)(pe-p), p, pe);
	    }
	    p = skipws(pe);
	    continue;
	 }
	 break;
      } while(1);

      if(!*p) break;

      my_dataset_item_arr_ensure(dataset->item, i+1);
      //if(!dataset->item[i].data) printf("new %d at %Ld\n", i, timestamp);
      my_data_arr_ensure(dataset->item[i].data, lineidx + 1);

      // TODO remove this hack
      if(num >= 84.9999999 && num <= 85.0000001) { num = NAN; } else
      {
	 // check minmax
	 if(dataset->item[i].max < num) dataset->item[i].max = num;
	 if(dataset->item[i].min > num) dataset->item[i].min = num;

	 // check diff minmax
	 if(lineidx > 0) {
	    double diff = (num - prev[i]) * TIME_MULTIPLIER / (double)cur_time_diff;
	    if(diff > 0) {
	       if(dataset->item[i].min_diff < diff) dataset->item[i].min_diff = diff;
	    } else {
	       if(dataset->item[i].max_diff < diff) dataset->item[i].max_diff = diff;
	    }
	 }
      }

      dataset->item[i].data[lineidx] = num;
      prev[i] = num;
   }
}

#define MAX_LINE_LEN 500

my_dataset_t *read_dataset_int(const char *file, int maxlines, progress_cb_t *cb, void *user_data) {
   FILE *f = fopen(file, "rt");
   if(f == NULL) return NULL;

   my_dataset_t *dataset = zmalloc_c(sizeof(my_dataset_t));

   dataset->item_dflt.min = INFINITY;
   dataset->item_dflt.max = -INFINITY;
   dataset->item_dflt.data_dflt = NAN;

   file = dataset->path = strdup(file);

   const char *filename = strrchr(file, '/');
   if(filename) {
      // slash found, jump to next letter
      filename++;
   } else {
      // no slash, use full name
      filename = file;
   }

   dataset->name = filename;

   time_t pd_l1 = time(0);

   char line[MAX_LINE_LEN];
   char *labels[100]; //##TODO## non-fixed
   int numlabels = 0;

   // fseek(f, -55 * 10000L, SEEK_END); fgets(line, (int)sizeof(line), f); // skip first partial line

   my_time_t ts;

   do {
      if(maxlines >= 0) {
	 if(maxlines-- == 0) {
	    fclose(f);
            int i;
	    for(i=0; i<numlabels; ++i) {
	       free(labels[i]);
	    }
	    free_dataset(dataset);
	    return NULL;
	 }
      }
      if(!fgets(line, (int)sizeof(line), f)) {
	 fclose(f);

         int i;
	 for(i=0; i<numlabels; ++i) {
	    free(labels[i]);
	 }
	 free_dataset(dataset);
	 return NULL;
      }

      if(numlabels == 0) {
	 // try to detect column labels
	 const char *x = skipws(line);
	 if(*x == '#') {
	    x++;
	    x = skipws(x);
	 }

	 // this "detection" is too simple but works in basic cases
	 if(isalpha(*x)) {
	    // labels found -> detect separator - check for tabs, multiple spaces or default to single space
	    const char *sep;
	    if(strchr(x, '\t')) {
	       sep = "\t";
	    } else if(strstr(x, "  ")) {
	       sep = "  ";
	    } else {
	       sep = " ";
	    }

	    char *next;
	    while((next = strstr(x, sep))) {
	       labels[numlabels++] = strnzdup(x, (size_t)(next - x));
	       x = skipws(next+1);
	    }
	    labels[numlabels++] = strdup(x);

	    continue;
	 }
      }

      strip_comments(line);

      const char *p = line;
      ts = parse_timestamp(&p);
      if(p == line || (!my_isspace(*p) && *p)) {
	 chomp(line);
	 fprintf(stderr, "Ignoring bad line '%s'\n", line);
	 continue;
      }

      break;
   } while(1);

   my_time_t timeoff = 0;

   dataset->timetype = TIMETYPE_EPOCH;
   dataset->ts = ts;
   boolean_t is_millis = 0;

   if(ts > 100000000000LL * TIME_MULTIPLIER) {
      is_millis = 1;

   } else if(ts < 100000000LL * TIME_MULTIPLIER) {
      dataset->timetype = TIMETYPE_START;
   }

   // get last modification time and assume that's also the time of the last entry
   struct stat st;
   if(stat(file, &st)) {
      perror("read_dataset(): failed to stat file");
   }

   dataset->last_modified = st.st_mtime * TIME_MULTIPLIER;

   // try to get the relative time of the last entry
   if(dataset->timetype == TIMETYPE_START) {
      long origoff = ftell(f);

      fseek(f, -MAX_LINE_LEN - 1L, SEEK_END);
      char eline[MAX_LINE_LEN];
      my_time_t te = 0;
      fgets(eline, (int)sizeof(eline), f); // skip first partial line
      while(fgets_strip_comments(eline, (int)sizeof(eline), f)) {
	 const char *p = eline;
	 my_time_t tmp = parse_timestamp(&p);
	 if(p > eline) {
	    // update only if successful
	    te = tmp;
	 }
      }

      timeoff = st.st_mtime * TIME_MULTIPLIER - te;

      dataset->ts += timeoff;

      fseek(f, origoff, SEEK_SET);
   }

   dataset->timeoff = timeoff;
   dataset->is_millis = is_millis;

   if(maxlines >= 0) {
      // scan mode.. load only part of data

      if(maxlines == 0) {
	 maxlines = 1;
      }

      int i;
      for(i=0; i<maxlines; ++i) {
	 parse_line(dataset, line, timeoff, is_millis);

	 if(!fgets_strip_comments(line, (int)sizeof(line), f)) break;
      }

      int origcount = dataset->time_count;

      // scan end of file also in case there are more data columns there

      fseek(f, -MAX_LINE_LEN - 1L, SEEK_END);
      fgets(line, (int)sizeof(line), f); // skip first partial line
      while(fgets_strip_comments(line, (int)sizeof(line), f)) {
	 parse_line(dataset, line, timeoff, is_millis);
      }

      dataset->te = dataset->time[dataset->time_count-1];

      // restore time count
      dataset->time_count = origcount;

   } else {
      // load whole file

      int idx = 0;

      while(1) {
	 parse_line(dataset, line, timeoff, is_millis);

	 if(unlikely(!(idx % 5000))) {
	    if(!(idx % 20000)) {
	       if(cb) {
		  cb(((double)ftell(f)) / st.st_size, user_data);
	       }
	    }

	    if(system_check_events()) {
	       // quit
	       dataset->item_count = 0;
	       break;
	    }
	 }
	 idx++;

	 if(unlikely(!fgets(line, (int)sizeof(line), f))) {
	    dataset->filepos = ftell(f);
	    break;
	 }

	 // ensure complete line
	 if(unlikely(!strrchr(line, '\n'))) {
	    dataset->filepos = ftell(f) - strlen(line);
	    break;
	 }

	 strip_comments(line);
      }


      dataset->te = dataset->time[dataset->time_count-1];
   }

   fclose(f);

   // make sure all items have same amount of elements and insert labels
   int i;
   for(i=0; i<dataset->item_count; ++i) {
      my_data_arr_ensure(dataset->item[i].data, dataset->time_count);
      if(i < numlabels-1) {
	 dataset->item[i].name = labels[i+1];
      }
   }

   if(dataset->item_count == 0) {
      free_dataset(dataset);
      return NULL;
   }

   time_t pd_l2 = time(0);
   if(!maxlines) printf("  loading time: %ld seconds\n", pd_l2 - pd_l1);

   ref_dataset(dataset);

   return dataset;
}

// return false if there was a problem TODO improve error reporting
boolean_t update_dataset(my_dataset_t *dataset) {
   // read more data from file

   struct stat st;
   if(stat(dataset->path, &st)) {
      perror(dataset->path);
      return 0;
   }

   if(st.st_size < dataset->filepos) {
      printf("File %s shrunk - update aborted\n", dataset->path);
      return 0;
   }

   FILE *f = fopen(dataset->path, "rt");
   if(f == NULL) {
      perror(dataset->path);
      return 0;
   }

   if(fseek(f, dataset->filepos, SEEK_SET)) {
      perror(dataset->path);
      fclose(f);
      return 0;
   }

   char line[MAX_LINE_LEN];
   my_time_t timeoff = dataset->timeoff;
   boolean_t is_millis = dataset->is_millis;
   int idx = 0;

   while(1) {
      if(unlikely(!fgets(line, (int)sizeof(line), f))) {
	 dataset->filepos = ftell(f);
	 break;
      }

      // ensure complete line
      if(unlikely(!strchr(line, '\n'))) {
	 dataset->filepos = ftell(f) - strlen(line);
	 break;
      }

      strip_comments(line);

      parse_line(dataset, line, timeoff, is_millis);

      if(unlikely(!(idx % 5000))) {
	 if(system_check_events()) {
	    fclose(f);
	    return 0;
	 }
      }
      idx++;
   }

   dataset->te = dataset->time[dataset->time_count-1];

   fclose(f);

   // make sure all items have same amount of elements and insert labels
   int i;
   for(i=0; i<dataset->item_count; ++i) {
      my_data_arr_ensure(dataset->item[i].data, dataset->time_count);
   }

   if(dataset->item_count == 0) {
      return 0;
   }

   return 1;
}

my_dataset_t *read_dataset_preview(const char *file, int maxsamples) {
   FILE *f = fopen(file, "rt");
   if(f == NULL) return NULL;

   my_dataset_t *dataset = zmalloc_c(sizeof(my_dataset_t));

   dataset->item_dflt.min = INFINITY;
   dataset->item_dflt.max = -INFINITY;
   dataset->item_dflt.data_dflt = NAN;

   file = dataset->path = strdup(file);

   const char *filename = strrchr(file, '/');
   if(filename) {
      // slash found, jump to next letter
      filename++;
   } else {
      // no slash, use full name
      filename = file;
   }

   dataset->name = filename;
   dataset->timetype = TIMETYPE_EPOCH;

   // get last modification time and assume that's also the time of the last entry
   struct stat st;
   if(stat(file, &st)) {
      perror("read_dataset(): failed to stat file");
   }

   dataset->last_modified = st.st_mtime * TIME_MULTIPLIER;

   char line[MAX_LINE_LEN];
   
   int i;
   for(i=0; i<maxsamples; ++i) {
      if(!fgets_strip_comments(line, (int)sizeof(line), f)) break;
      parse_line(dataset, line, 0LL, 0);
      int off = st.st_size / (maxsamples + 1) * i - ftell(f);
      if(off > 0)
	 fseek(f, off, SEEK_CUR);

      // skip partial line
      if(!fgets(line, (int)sizeof(line), f)) break;
   }

   fclose(f);

   dataset->ts = dataset->time[0];
   dataset->te = dataset->time[dataset->time_count-1];
   dataset->maxdiff = dataset->te - dataset->ts;

   // make sure all items have same amount of elements and insert labels
   for(i=0; i<dataset->item_count; ++i) {
      my_data_arr_ensure(dataset->item[i].data, dataset->time_count);
   }

   if(dataset->item_count == 0) {
      free_dataset(dataset);
      return NULL;
   }

   ref_dataset(dataset);

   return dataset;
}

/*
void add_dataset_listener(my_dataset_t *dataset, my_dataset_update_cb_t *cb, void *user_data) {
   
}

void remove_dataset_listener(my_dataset_t *dataset, my_dataset_update_cb_t *cb, void *user_data) {
   
}
*/

static void free_dataset(my_dataset_t *dataset) {
   int i;
   for(i=0; i<dataset->item_count; ++i) {
      my_data_arr_free(dataset->item[i].data);
      if(dataset->item[i].name) {
	 free(dataset->item[i].name);
      }
   }

   my_dataset_item_arr_free(dataset->item);

   if(dataset->path) {
      free(dataset->path);
   }

   bzero(dataset, sizeof(my_dataset_t));
   free(dataset);
}

void ref_dataset(my_dataset_t *dataset) {
   ++dataset->refcount;
}

int unref_dataset(my_dataset_t *dataset) {
   if(--dataset->refcount <= 0) {
      //printf("Freeing dataset %s\n", dataset->name);
      free(dataset);
      return 1;
   }
   return 0;
}
