/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2 - most functions related to representing & managing the
 * actual graphs along with gtk event handlers - except for the
 * plotting itself, see plot_draw
 */

#include <gtk/gtkbox.h>
#include <gtk/gtktogglebutton.h>
#include <math.h>

#include "../system.xh"
#include "compat.xh"
#include "eventhandlers.xh"
#include "paintapi-gtk.xh"
#include "plot2.xh"
#include "tail_fam.xh"

// ##TODO## make local
// ##TODO## the colors suck
paintapi_rgb_t default_line_colors[] = {

   { 0xffff, 0x4444, 0x4444 },
   { 0xffff, 0xc0c0, 0x4040 },
   { 0x4040, 0xffff, 0x4040 },
   { 0x4040, 0xffff, 0xffff },
   { 0x4040, 0x8080, 0xffff },
   { 0xc0c0, 0x4040, 0xffff },
   { 0xc0c0, 0xc0c0, 0xc0c0 },

   { 0x8080, 0x2222, 0x2222 },
   { 0x8080, 0x6060, 0x2020 },
   { 0x2020, 0x8080, 0x2020 },
   { 0x2020, 0x8080, 0x8080 },
   { 0x2020, 0x4040, 0x8080 },
   { 0x6060, 0x2020, 0x8080 },
   { 0x6060, 0x6060, 0x6060 },

};

int num_default_line_colors = SZ(default_line_colors);

void remove_data_file(my_plot_t *plot, int idx) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_remove_data_file(&plot->view[vi]->draw, idx);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

void remove_all_data_files(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_remove_all_data_files(&plot->view[vi]->draw);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

void reset_gc(my_plot_t *plot) {
   GdkColor bg;
   bg.red = bg.green = bg.blue = plot->view[0]->draw.inverse_colors ? 65535 : 0;
   gdk_colormap_alloc_color(gdk_colormap_get_system(), &bg, 0, 0);

   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_reset_gc(&plot->view[vi]->draw);
      gdk_window_set_background(plot->view[vi]->drawable, &bg);
   }
}

void refresh_all_files(my_plot_t *plot) {
   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      update_dataset(plot->view[0]->draw.line_set[lsi].dataset);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

void plot_init(my_plot_t *plot) {
   bzero(plot, sizeof(my_plot_t));

   // init default view
   plot->view_dflt.parent = plot; // back ref
   plot->view_dflt.hline = -1;
   plot->view_dflt.vline = -1;
   plot_draw_init(&plot->view_dflt.draw);
}

void plot_free(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      plot_draw_deinit(&plot->view[vi]->draw);
      if(plot->view[vi]->draw.font_legend) {
	 free(plot->view[vi]->draw.font_legend);
      }
   }

   data_scrollbar_controller_free(plot->dsc_x);
   for(int vi=0; vi<plot->view_count; ++vi) {
      data_scrollbar_controller_free(plot->view[0]->dsc_y);
   }

   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->pangofont) {
	 pango_font_description_free(plot->view[vi]->pangofont);
      }
      if(plot->view[vi]->pangoctx && !plot->view[vi]->da) {
	 g_free(plot->view[vi]->pangoctx);
      }
   }

#ifdef USE_FAM
   if(plot->fam_initialized) {
      g_source_remove(plot->fam_evtsrcid);
      FAMClose(&plot->fc);
   }
#endif

   bzero(plot, sizeof(my_plot_t));
   free(plot);
}

void plot_view_clone(my_plot_view_t *n, my_plot_view_t *o) {
   n->draw.inverse_colors = o->draw.inverse_colors;
   n->draw.samples_per_pixel = o->draw.samples_per_pixel;
   n->draw.show_legend = o->draw.show_legend;
   n->draw.show_disabled = o->draw.show_disabled;

   data_scrollbar_controller_copy(n->dsc_y, o->dsc_y);

   n->draw.xstart = o->draw.xstart;
   n->draw.xend = o->draw.xend;
   n->draw.xmin = o->draw.xmin;
   n->draw.xmax = o->draw.xmax;
   n->draw.ymin = o->draw.ymin;
   n->draw.ymax = o->draw.ymax;

   n->draw.minor_color = o->draw.minor_color;
   n->draw.major_color = o->draw.major_color;
   n->draw.zerogrid_color = o->draw.zerogrid_color;
   n->draw.sep_color = o->draw.sep_color;
   n->draw.label_color = o->draw.label_color;
   n->draw.cursor_marker_color = o->draw.cursor_marker_color;
   n->draw.legend_bg_color = o->draw.legend_bg_color;
   n->draw.legend_border_color = o->draw.legend_border_color;
   n->draw.legend_text_color = o->draw.legend_text_color;

   int setcount = o->draw.line_set_count;
   my_plot_line_info_set_arr_ensure(n->draw.line_set, setcount);

   for(int lsi = 0; lsi < setcount; ++lsi) {
      my_plot_line_info_set_t *ns, *os;
      ns = &n->draw.line_set[lsi];
      os = &o->draw.line_set[lsi];

      ns->dataset = os->dataset;
      ref_dataset(os->dataset);

      int linecount = os->line_count;
      my_plot_line_info_arr_ensure(ns->line, linecount);

      for(int li = 0; li < linecount; ++li) {
	 my_plot_line_info_t *nsl, *osl;
	 nsl = &ns->line[li];
	 osl = &os->line[li];

	 nsl->data = osl->data;
	 nsl->color = osl->color;
	 nsl->dataset_idx = osl->dataset_idx;
	 nsl->enabled = osl->enabled;
	 nsl->mode = osl->mode;
      }
   }
}

my_plot_t *plot_clone(my_plot_t *o) {
   my_plot_t *n = malloc_c(sizeof(my_plot_t));
   plot_init(n);

   my_plot_view_arr_ensure(n->view, o->view_count);

   n->dsc_x = data_scrollbar_controller_clone(o->dsc_x);
   n->fontsize_at_width = o->fontsize_at_width;
   n->scale_font_by_size = o->scale_font_by_size;

   for(int vi=0; vi<o->view_count; ++vi) {
      n->view[vi]->dsc_y = data_scrollbar_controller_clone(o->view[vi]->dsc_y);
      plot_view_clone(n->view[vi], o->view[vi]);
   }

   return n;
}

static my_plot_view_t *plot_get_prio_view(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->bmode) {
	 return plot->view[vi];
      }
   }
   return NULL;
}

void plot_need_redraw(my_plot_t *plot) {
   //printf("need_redraw\n");
   my_plot_view_t *prio = plot_get_prio_view(plot);
   if(prio) {
      // some view is prioritized, render it and keep the others waiting
      plot_view_need_redraw(prio);
      for(int vi=0; vi<plot->view_count; ++vi) {
	 if(plot->view[vi] == prio) continue;
	 plot->view[vi]->normal_expose++;
      }
   } else {
      for(int vi=0; vi<plot->view_count; ++vi) {
	 plot_view_need_redraw(plot->view[vi]);
      }
   }
}

void plot_view_need_redraw(my_plot_view_t *view) {
   //printf("view_need_redraw\n");
   if(view->conc == 0) {
      if(view->da) gtk_widget_queue_draw(view->da);
   } else {
      view->normal_expose++;
   }
}

void plot_size_or_font_changed(my_plot_t *plot) {
   int w,h;
   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      gdk_drawable_get_size(view->drawable, &w, &h);

      PangoFontDescription *font_desc;
      if(plot->scale_font_by_size) {
	 int fontlen = strlen(plot->font);
	 char *font = malloc((size_t)(fontlen + 40));
	 char *space = strrchr(plot->font, ' ');
	 double size;
	 if(!space) {
	    space = font + fontlen;
	    size = 12;
	 } else {
	    size = atof(space+1);
	 }
	 memcpy(font, plot->font, (size_t)(space - plot->font));

	 // scrollbars etc take 72x110 pixels
	 // good fontsizes:
	 // at 640x480, wh = 530.66 native-fontsize = 9 scaled-fontsize=11
	 // at 640x480, wh = 640 native-fontsize = 10
	 // at 1600x1200, wh = 1491.33 native-fontsize = 17
	 // at 1600x1200, wh = 1600 native-fontsize = 18
	 // first scale according to screen size

	 // scaled = 2.4^(x/1600-1) * 18.3
	 // native = 3.25^(x/1600-1) * 18.3

	 // 18.3 * pow(1.4,sw/1600-1) * pow(2.15, ((pw - 72 + (pw*3/4-110)*4/3)/2)/1600 -1 )

	 //double x1 = size;

	 int cur_screen_width = plot->view[0]->da ? gdk_screen_get_width(gtk_window_get_screen(plot->parent)) : 1600;

	 size *= pow(1.4, ((double)cur_screen_width) / 1600 - 1);

	 //double x2 = size;

	 // then scale according to widget size
	 double wh = (w + 4.0 * h / 3.0) / 2.0;

	 //size *= exp2(wh / cur_screen_width - 1);
	 //size *= exp2(wh / plot->fontsize_at_width - 1);
	 size *= pow(2.15, wh / 1600 - 1);

	 snprintf(font + (space-plot->font), (size_t)40, " %.1lf", round(size * 2) / 2);

	 //printf("font %s / %lf -> %lf -> %lf   wh %lf  %dx%d   (%d)\n", font, x1, x2, size, wh, w, h, cur_screen_width);

	 font_desc = pango_font_description_from_string(font);

	 free(font);

      } else {
	 font_desc = pango_font_description_from_string(plot->font);
      }

      if(view->pangofont) {
	 pango_font_description_free(view->pangofont);
      }

      view->pangofont = font_desc;

      if(!view->pangoctx) {
	 if(view->da) {
	    view->pangoctx = gtk_widget_get_pango_context(view->da);
	 } else {
	    view->pangoctx = gdk_pango_context_get();
	 }
      }

      // update plot_draw
      if(!view->draw.font_legend) {
	 view->draw.font_legend = zmalloc_c(sizeof(paintapi_font_t));
	 view->draw.font_xaxis = view->draw.font_legend;
	 view->draw.font_yaxis = view->draw.font_legend;
      }
      view->draw.font_legend->ctx = view->pangoctx;
      view->draw.font_legend->font = view->pangofont;

      view->draw.plot_xe = w - 1;
      view->draw.plot_ye = h - 1;
   }

   plot_need_redraw(plot);
}

static void atzh_helper(my_zoom_view_t *z, data_scrollbar_controller_t *dsc) {
   z->off = data_scrollbar_controller_get_page_off(dsc);
   z->length_log = data_scrollbar_controller_get_page_length_log(dsc);
   z->scalepos = data_scrollbar_controller_get_zoompoint(dsc);
}

void add_to_zoom_history(my_plot_t *plot) {
   if(plot->view[0]->draw.line_set_count == 0) {
      // don't save history if no plots
      return;
   }

   int idx = plot->zoom_pos + 1;
   if(idx > plot->zoom_history_count) idx = plot->zoom_history_count;

   my_zoom_view_arr_ensure(plot->zoom_history, idx + 1);
   atzh_helper(&plot->zoom_history[idx], plot->dsc_x);

   for(int vi=0; vi<plot->view_count; ++vi) {
      atzh_helper(&plot->view[vi]->zoom_history_dflt, plot->view[vi]->dsc_y); // TODO do this in clone instead and get rid of CUSTOM default
      my_zoom_view_arr_ensure(plot->view[vi]->zoom_history, idx + 1);
      atzh_helper(&plot->view[vi]->zoom_history[idx], plot->view[vi]->dsc_y);
   }

   if(idx > 0) {
      gtk_widget_set_sensitive(plot->button_zoom_prev, 1);
   }
   gtk_widget_set_sensitive(plot->button_zoom_next, 0);

   plot->zoom_pos = idx;
}

void show_all_nohistory(my_plot_view_t *view) {
   double min, max;

   data_scrollbar_controller_get_range(view->parent->dsc_x, &min, &max);
   data_scrollbar_controller_set_page_length(view->parent->dsc_x, max - min);

   data_scrollbar_controller_get_range(view->dsc_y, &min, &max);
   data_scrollbar_controller_set_page_length(view->dsc_y, max - min);
}

void show_all(my_plot_view_t *view) {
   show_all_nohistory(view);
   add_to_zoom_history(view->parent);
}

void pan_right(my_plot_t *plot) {
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

   double min, max;
   data_scrollbar_controller_get_range(plot->dsc_x, &min, &max);

   data_scrollbar_controller_set_page_off(plot->dsc_x, max - xlength);
}

void switch_to_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;
   if(idx == plot->zoom_pos) return;

   plot->zoom_pos = idx;

   gtk_widget_set_sensitive(plot->button_zoom_prev, idx > 0);
   gtk_widget_set_sensitive(plot->button_zoom_next, idx < plot->zoom_history_count-1);

   my_zoom_view_t *z = &plot->zoom_history[idx];

   data_scrollbar_controller_set_page_length_log(plot->dsc_x, z->length_log);
   data_scrollbar_controller_set_page_off(plot->dsc_x, z->off);
   data_scrollbar_controller_set_zoompoint(plot->dsc_x, z->scalepos);

   for(int vi=0; vi<plot->view_count; ++vi) {
      z = &plot->view[vi]->zoom_history[idx];

      data_scrollbar_controller_set_page_length_log(plot->view[vi]->dsc_y, z->length_log);
      data_scrollbar_controller_set_page_off(plot->view[vi]->dsc_y, z->off);
      data_scrollbar_controller_set_zoompoint(plot->view[vi]->dsc_y, z->scalepos);
   }
}

char cursor_marker_dashes[2] = { 2, 4 };

void show_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;

   my_zoom_view_t *z = &plot->zoom_history[idx];

   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

   double z__xoff2 = z->off + exp(z->length_log);

   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      z = &view->zoom_history[idx];

      double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);
      double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);

      double z__yoff2 = z->off + exp(z->length_log);

      int x = (int)(view->draw.plot_xe * (z->off - xoff) / xlength);
      int y = (int)(view->draw.plot_yde * (z->off - yoff) / ylength);
      int x2 = (int)(view->draw.plot_xe * (z__xoff2 - xoff) / xlength);
      int y2 = (int)(view->draw.plot_yde * (z__yoff2 - yoff) / ylength);

      y = view->draw.plot_yde - y;
      y2 = view->draw.plot_yde - y2;

      view->bx = x;
      view->by = y;
      view->box = x2;
      view->boy = y2;

      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_rectangle(view->draw.paintapi, view->draw.cursor_marker_gc, 0, x, y, x2, y2);
   }
}

void hide_zoom_history(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      if(plot->view[vi]->box == COORD_UNDEF) return;

      plot->view[vi]->draw.paintapi->gc_set_dashes(plot->view[vi]->draw.paintapi, plot->view[vi]->draw.cursor_marker_gc, -plot->view[vi]->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      plot->view[vi]->draw.paintapi->draw_rectangle(plot->view[vi]->draw.paintapi, plot->view[vi]->draw.cursor_marker_gc, 0, plot->view[vi]->bx, plot->view[vi]->by, plot->view[vi]->box, plot->view[vi]->boy);

      plot->view[vi]->box = COORD_UNDEF;
   }
}

int plot_add_view_with_xml(my_plot_t *plot, GladeXML *gladexml) {
   int idx = plot->view_count;
   my_plot_view_arr_ensure(plot->view, idx+1);
   my_plot_view_t *view = plot->view[idx];

   view->da = glade_xml_get_widget(gladexml, "drawingarea1");
   view->dsc_y = data_scrollbar_controller_new(gladexml, "yscroll_");
   view->box_yscrolls_plot = glade_xml_get_widget(gladexml, "box_yscrolls_plot");;

   plot_view_connect_signals(view, gladexml);

   return idx;
}

void plot_add_view(my_plot_t *plot) {
   GladeXML *gladexml = glade_xml_new(plot->gladefile, "box_yscrolls_plot", NULL);
   GtkWidget *box = glade_xml_get_widget(gladexml, "box_yscrolls_plot");
   gtk_box_pack_start(GTK_BOX(plot->box_plots), box, 1, 1, 0);

   int idx = plot_add_view_with_xml(plot, gladexml);

   plot_view_clone(plot->view[idx], plot->view[idx-1]);

   update_view_minmax(plot->view[idx]);
}

void show_view_markers(my_plot_view_t *view, boolean_t show) {
   if(view->hmarkers_drawn != show && view->hmarkers_set) {
      view->hmarkers_drawn = show;
      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.cursor_marker_gc, 0, view->hline, view->draw.plot_xe, view->hline);
   }
   if(view->vmarkers_drawn != show && view->vmarkers_set) {
      view->vmarkers_drawn = show;
      view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -view->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.cursor_marker_gc, view->vline, 0, view->vline, view->draw.plot_yde);
   }

   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 if(view->draw.line_set[lsi].line[li].markers_drawn != show && view->draw.line_set[lsi].line[li].hline != -1) {
	    view->draw.line_set[lsi].line[li].markers_drawn = show;
	    view->draw.paintapi->draw_line(view->draw.paintapi,
					   view->draw.line_set[lsi].line[li].horiz_marker,
					   0, view->draw.line_set[lsi].line[li].hline,
					   view->draw.plot_xe, view->draw.line_set[lsi].line[li].hline);
	 }
      }
   }
}

void show_markers(my_plot_t *plot, boolean_t show) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_view_markers(plot->view[vi], show);
   }
}

void reset_view_markers(my_plot_view_t *view) {
   view->hmarkers_drawn = 0;
   view->vmarkers_drawn = 0;
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 view->draw.line_set[lsi].line[li].markers_drawn = 0;
      }
   }
}

void reset_markers(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      reset_view_markers(plot->view[vi]);
   }
}

void destroy_view_markers(my_plot_view_t *view) {
   show_view_markers(view, 0);
   view->hmarkers_set = 0;
   view->vmarkers_set = 0;
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 view->draw.line_set[lsi].line[li].hline = -1;
      }
   }
}

void destroy_markers(my_plot_t *plot) {
   for(int vi=0; vi<plot->view_count; ++vi) {
      destroy_view_markers(plot->view[vi]);
   }
}

void set_and_show_markers(my_plot_view_t *view, int x, int y) {
   if(!view->draw.minor_gc) {
      // gui not yet initialized
      return;
   }

   destroy_markers(view->parent);

   if(y > view->draw.plot_yde) return;

   view->draw.paintapi->gc_set_dashes(view->draw.paintapi, view->draw.cursor_marker_gc, -x, cursor_marker_dashes, SZ(cursor_marker_dashes));
   view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.cursor_marker_gc, 0, y, view->draw.plot_xe, y);
   view->hmarkers_drawn = 1;
   view->hmarkers_set = 1;

   int yo = 0;
   for(int vi=0; vi<view->parent->view_count; ++vi) {
      my_plot_view_t *v2 = view->parent->view[vi];
      v2->draw.paintapi->gc_set_dashes(v2->draw.paintapi, v2->draw.cursor_marker_gc, -y, cursor_marker_dashes, SZ(cursor_marker_dashes));
      v2->draw.paintapi->draw_line(v2->draw.paintapi, v2->draw.cursor_marker_gc, x, 0, x, v2->draw.plot_yde);
      v2->hline = y;
      v2->vline = x;
      v2->vmarkers_drawn = 1;
      v2->vmarkers_set = 1;
      yo += v2->draw.plot_ye + 1;
   }

   double xlength = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(view->parent->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(view->dsc_y);

   my_time_t xv = (my_time_t)(x * xlength / view->draw.plot_xe + xoff);
   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      int p = my_time_arr_find(view->draw.line_set[lsi].dataset->time, xv, 0);
      for(int li=0; li<view->draw.line_set[lsi].line_count; ++li) {
	 if(!view->draw.line_set[lsi].line[li].enabled) continue;

	 int dataset_idx = view->draw.line_set[lsi].line[li].dataset_idx;

	 int dy = !__isnan(view->draw.line_set[lsi].dataset->item[dataset_idx].data[p]) ?
	    (int)(view->draw.plot_yde - (view->draw.line_set[lsi].dataset->item[dataset_idx].data[p] - yoff) / ylength * view->draw.plot_yde) :
	    -1;

	 if(dy >= 0) {
	    view->draw.paintapi->draw_line(view->draw.paintapi, view->draw.line_set[lsi].line[li].horiz_marker, 0, dy, view->draw.plot_xe, dy);
	    view->draw.line_set[lsi].line[li].hline = dy;
	    view->draw.line_set[lsi].line[li].markers_drawn = 1;
	 }
      }
   }
}

void plot_move_to(my_plot_view_t *view, int x, int y) {

/*
  data_scrollbar_controller_set_zoompoint(view->dsc_x, ((double)x) / view->draw.plot_xe);
  data_scrollbar_controller_set_zoompoint(view->dsc_y, 1 - ((double)y) / view->draw.plot_yde);
*/

   int xd = x - view->bx;
   int yd = view->by - y;

   xd *= view->bspeed;
   yd *= view->bspeed;

   //printf("move %d,%d\n", xd, yd);

   double xmin, xmax;
   double ymin, ymax;

   data_scrollbar_controller_get_range(view->parent->dsc_x, &xmin, &xmax);
   data_scrollbar_controller_get_range(view->dsc_y, &ymin, &ymax);

   double xlength = data_scrollbar_controller_get_page_length(view->parent->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(view->dsc_y);

   double nxoff = view->bxoff - xd * xlength / view->draw.plot_xe;
   double nyoff = view->byoff - yd * ylength / view->draw.plot_yde;

   //printf("nt %lf %lf\n", nxoff, nyoff);

   if(nxoff < xmin) nxoff = xmin; else if(nxoff + xlength > xmax) nxoff = xmax - xlength;
   if(nyoff < ymin) nyoff = ymin; else if(nyoff + ylength > ymax) nyoff = ymax - ylength;

   //printf("n %lf %lf\n", nxoff, nyoff);

   data_scrollbar_controller_set_page_off(view->parent->dsc_x, nxoff);
   data_scrollbar_controller_set_page_off(view->dsc_y, nyoff);
}


#define VDIFFMIN_DIVISOR 1000

void update_view_minmax(my_plot_view_t *view) {
   double vmin = INFINITY, vmax = -INFINITY;
   double vdiffmin = VDIFFMIN_DIVISOR;

   int count = 0;

   for(int lsi=0; lsi<view->draw.line_set_count; ++lsi) {
      my_dataset_t *dataset = view->draw.line_set[lsi].dataset;
      my_plot_line_info_t *line = view->draw.line_set[lsi].line;
      int line_count = view->draw.line_set[lsi].line_count;

      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 int d = line[li].dataset_idx;
	 switch(line[li].mode) {
	    case LINE_MODE_NORMAL: {
	       if(count == 0 || vmin > dataset->item[d].min) vmin = dataset->item[d].min;
	       if(count == 0 || vmax < dataset->item[d].max) vmax = dataset->item[d].max;

	       // TODO should we include disabled lines in vdiffmin calculation...
	       double vdiff = dataset->item[d].max - dataset->item[d].min;
	       if(count == 0 || vdiffmin > vdiff) vdiffmin = vdiff;

	       break;
	    }
	    case LINE_MODE_DIFF: {
	       if(count == 0 || vmin > dataset->item[d].min_diff) vmin = dataset->item[d].min_diff;
	       if(count == 0 || vmax < dataset->item[d].max_diff) vmax = dataset->item[d].max_diff;

	       // TODO see above
	       double vdiff = dataset->item[d].max_diff - dataset->item[d].min_diff;
	       if(count == 0 || vdiffmin > vdiff) vdiffmin = vdiff;

	       break;
	    }
	 }
	 count++;
      }
   }

   // don't do anything if no lines visible
   if(count == 0) return;

   data_scrollbar_controller_set_range(view->dsc_y, vmin - (vmax-vmin)/20, vmax + (vmax-vmin)/20);

   //data_scrollbar_controller_set_page_length(plot->dsc_y, vmax-vmin); // initially zoom out

   // TODO maybe use diff data as base for minimum length instead of a fraction of the full range

   double yminlength = vdiffmin > 0 ? vdiffmin / VDIFFMIN_DIVISOR : 1;
   data_scrollbar_controller_set_min_interval(view->dsc_y, yminlength);
}

void update_x_minmax(my_plot_t *plot) {
   my_time_t ts = 0, te = 1;
   my_time_t min_time_diff = 1;

   int lsi;
   for(lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->view[0]->draw.line_set[lsi].dataset;

      if(lsi == 0 || min_time_diff > dataset->min_time_diff) {
	 min_time_diff = dataset->min_time_diff;
      }

      if(lsi == 0 || dataset->ts < ts) {
	 ts = dataset->ts;
      }

      if(lsi == 0 || dataset->te > te) {
	 te = dataset->te;
      }
   }

   // don't do anything if no datasets available
   if(lsi == 0) return;

   data_scrollbar_controller_set_range(plot->dsc_x, (double)ts, (double)te);
   data_scrollbar_controller_set_min_interval(plot->dsc_x, min_time_diff * 10.0);
}

void update_minmax(my_plot_t *plot) {
   update_x_minmax(plot);

   for(int vi=0; vi<plot->view_count; ++vi) {
      update_view_minmax(plot->view[vi]);
   }
}

int linecoloroff = 0;

static void ad_update_progress(double percent, void *plotp) {
   my_plot_t *plot = (my_plot_t*)plotp;

   gtk_progress_bar_set_fraction(plot->progressbar, percent);
}

void add_datafile(my_plot_t *plot, char *filename) {
   printf("Loading %s ..\n", filename);

   char tmp[40 + strlen(filename)];
   sprintf(tmp, "Loading %s ..", filename);
   gtk_widget_show(GTK_WIDGET(plot->progressbar));
   gtk_progress_bar_set_text(plot->progressbar, tmp);

   gtk_widget_set_sensitive(plot->menuitem_file_add_data_file, 0);

   my_dataset_t *dataset = read_dataset(filename, ad_update_progress, plot);

   gtk_widget_set_sensitive(plot->menuitem_file_add_data_file, 1);

   gtk_progress_bar_set_text(plot->progressbar, "");
   gtk_progress_bar_set_fraction(plot->progressbar, 0.0);
   gtk_widget_hide(GTK_WIDGET(plot->progressbar));

   if(!dataset) {
      if(!quit) {
	 fprintf(stderr, "\033[31;1mWARNING: File '%s' could not be loaded\033[m\n", filename);
      }
      return;
   }

   dataset->maxdiff = 100 * TIME_MULTIPLIER;

   int lsi = plot->view[0]->draw.line_set_count;
   for(int vi=0; vi<plot->view_count; ++vi) {
      my_plot_view_t *view = plot->view[vi];

      my_plot_line_info_set_arr_ensure(view->draw.line_set, lsi + 1);

      view->draw.line_set[lsi].dataset = dataset;

      if(vi == 0) printf("Datafile %d: %d data columns, %d samples\n", lsi + 1, dataset->item_count, dataset->time_count);

      for(int dsi=0; dsi<dataset->item_count; ++dsi) {
	 int li = view->draw.line_set[lsi].line_count;
	 my_plot_line_info_arr_ensure(view->draw.line_set[lsi].line, li+1);
	 int linecoloridx = linecoloroff + dsi;
	 view->draw.line_set[lsi].line[li].color = default_line_colors[linecoloridx % SZ(default_line_colors)];
	 view->draw.line_set[lsi].line[li].dataset_idx = dsi;
	 view->draw.line_set[lsi].line[li].enabled = 1;
	 view->draw.line_set[lsi].line[li].mode = LINE_MODE_NORMAL;
	 if(vi == 0) printf("  column %d: value range %lf - %lf\n", dsi+1, dataset->item[dsi].min, dataset->item[dsi].max);
      }
   }

   linecoloroff += dataset->item_count;

   puts("");

#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail));
   if(tailing) {
      enable_fam(plot);
   }
#endif

   plot_need_redraw(plot);
   update_minmax(plot);
   for(int vi=0; vi<plot->view_count; ++vi) {
      show_all_nohistory(plot->view[vi]);
   }

   if(lsi == 0) {
      // set initial center
      double min, max;
      for(int vi=0; vi<plot->view_count; ++vi) {
	 data_scrollbar_controller_get_range(plot->view[vi]->dsc_y, &min, &max);
	 data_scrollbar_controller_set_zoompoint(plot->view[vi]->dsc_y, min > 0 ? 0 : max < 0 ? 1 : -min / (max - min)); // align to(wards) zero
      }
      data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0); // align to right
   }

   add_to_zoom_history(plot);
}
