/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * tail_fam - support "tailing" using the FAM interface
 */

#ifdef USE_FAM

#include "tail_fam.xh"

#include <stdio.h>

static void FAMError(const char *msg) {
   fprintf(stderr, "%s: %s\n", msg, FamErrlist[FAMErrno]);
}

static gboolean fam_callback(GIOChannel *source, GIOCondition condition, my_plot_t *plot) {
   boolean_t updated = 0;
   while(FAMPending(&plot->fc)) {
      FAMEvent fe;
      //printf("FAM event\n");
      if(FAMNextEvent(&plot->fc, &fe) == -1) {
	 FAMError("fammonitorfile");
	 exit(1);
      }

      if(fe.code != FAMChanged) continue;

      //printf("FAM event changed\n");

      // TODO check only the file that the FAMChanged event indicates

      for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
	 if(!plot->view[0]->draw.line_set[lsi].dataset->fam_monitored) continue;
	 if(plot->view[0]->draw.line_set[lsi].dataset->fr.reqnum != fe.fr.reqnum) continue;
	 int oldcount = plot->view[0]->draw.line_set[lsi].dataset->time_count;
	 if(!update_dataset(plot->view[0]->draw.line_set[lsi].dataset)) {
	    printf("Problems updating %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
	    // TODO maybe ask user whether to start from beginning
	    printf("remove watch for %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
	    if(FAMCancelMonitor(&plot->fc, &plot->view[0]->draw.line_set[lsi].dataset->fr)) {
	       fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->view[0]->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
	    }
	    plot->view[0]->draw.line_set[lsi].dataset->fam_monitored = 0;
	 }
	 if(oldcount < plot->view[0]->draw.line_set[lsi].dataset->time_count) {
	    updated = 1;
	 }
      }
   }

   if(updated) {
      //printf("FAM updated\n");
      plot_need_redraw(plot);
      update_minmax(plot);
      pan_right(plot);
      // avoid zoompoint gliding implied by the zoompoint-maintaining algorithm in data_scrollbar_controller
      data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);
   }

   return 1;
}

void enable_fam(my_plot_t *plot) {
   if(!plot->fam_initialized) {
      if(FAMOpen(&plot->fc)) {
	 FAMError("fammonitorfile");
	 return;
      }

      int fd = FAMCONNECTION_GETFD(&plot->fc);

      GIOChannel *fam_giochannel = g_io_channel_unix_new(fd);
      if(!fam_giochannel) {
	 FAMClose(&plot->fc);
	 return;
      }
      plot->fam_evtsrcid = g_io_add_watch(fam_giochannel, G_IO_IN, (GIOFunc)fam_callback, plot);
      g_io_channel_unref(fam_giochannel);
      plot->fam_initialized = 1;
   }

   gtk_widget_set_sensitive(plot->xscroll_off, 0);
   gtk_widget_set_sensitive(plot->xscroll_off_rel, 0);
   gtk_widget_set_sensitive(plot->xscroll_zoompoint, 0);
   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 0);

   pan_right(plot);

   data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);

   char path[PATH_MAX];

   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      if(plot->view[0]->draw.line_set[lsi].dataset->fam_monitored) continue;
      realpath(plot->view[0]->draw.line_set[lsi].dataset->path, path);
      printf("add watch for %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
      if(FAMMonitorFile(&plot->fc, path, &plot->view[0]->draw.line_set[lsi].dataset->fr, plot)) {
	 fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->view[0]->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
      }
      plot->view[0]->draw.line_set[lsi].dataset->fam_monitored = 1;
   }
}

void disable_fam(my_plot_t *plot) {
   if(!plot->fam_initialized) return;

   gtk_widget_set_sensitive(plot->xscroll_off, 1);
   gtk_widget_set_sensitive(plot->xscroll_off_rel, 1);
   gtk_widget_set_sensitive(plot->xscroll_zoompoint, 1);
   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 1);

   for(int lsi=0; lsi<plot->view[0]->draw.line_set_count; ++lsi) {
      if(!plot->view[0]->draw.line_set[lsi].dataset->fam_monitored) continue;
      printf("remove watch for %s\n", plot->view[0]->draw.line_set[lsi].dataset->path);
      if(FAMCancelMonitor(&plot->fc, &plot->view[0]->draw.line_set[lsi].dataset->fr)) {
	 fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->view[0]->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
      }
      plot->view[0]->draw.line_set[lsi].dataset->fam_monitored = 0;
   }

   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 1);
}

#endif
